const ConsoleLogger = require('moleculer').Loggers.Console
const FormattedLogger = require('moleculer').Loggers.Formatted
const {AsyncLocalStorage} = require('async_hooks')

const asyncLocalStorage = new AsyncLocalStorage()
asyncLocalStorage.enterWith(new Map())

class Logger extends ConsoleLogger {
    getLogHandler(bindings) {
        const level = bindings ? this.getLogLevel(bindings.mod) : null
        if (!level) {
            return null
        }
        bindings.mod = bindings.mod ? bindings.mod : ''
        const levelIdx = FormattedLogger.LEVELS.indexOf(level)

        return (type, args) => {
            const typeIdx = FormattedLogger.LEVELS.indexOf(type)
            if (typeIdx > levelIdx) {
                return
            }
            const store = asyncLocalStorage.getStore()
            const log = JSON.stringify({
                type: type,
                requestId: store.get('requestId'),
                trace: args.reduce((r, v) => r += ` ${v}`, ''),
            })

            console.log(`|${log}|`)
        }
    }
}

function wrapper(err) {
    const store = asyncLocalStorage.getStore()
    const json = JSON.stringify({
        type: 'fatal',
        requestId: store.get('requestId'),
        trace: err.stack,
    })
    return `|${json}|`
}

function traceWrapper(responseTime, operation, type, params, startTime) {
    const store = asyncLocalStorage.getStore()
    const json = JSON.stringify({
        type: type,
        requestId: store.get('requestId'),
        operation: operation,
        responseTime: responseTime,
        startTime: startTime
    })
    return `|${json}|`
}

function isPromise(obj) {
    return !!obj && (typeof obj === 'object' || typeof obj === 'function') && typeof obj.then === 'function'
}

const logger = new Logger()

module.exports = {
    middleware: {
        name: 'Logger',
        localAction(next, action) {
            return async (ctx) => {
                const store = new Map()
                return asyncLocalStorage.run(store, async () => {
                    store
                        .set('requestId', ctx.requestID)
                    const start = Date.now()
                    try {
                        const result = await next(ctx)
                        const responseTime = Date.now() - start
                        console.log(traceWrapper(responseTime, action.name, 'action', ctx.params, start))
                        return result
                    } catch (e) {
                        console.error(wrapper(e))
                        throw e
                    }
                })
            }
        },
        localEvent(next, event) {
            return async (ctx) => {
                const store = new Map()
                return asyncLocalStorage.run(store, async () => {
                    store
                        .set('requestId', ctx.requestID)
                    const start = Date.now()
                    try {
                        const result = await next(ctx)
                        const responseTime = Date.now() - start
                        if (!event.name.includes('$')) {
                            console.log(traceWrapper(responseTime, event.name, 'event', ctx.params, start))
                        }
                        return result
                    } catch (e) {
                        console.error(wrapper(e))
                        throw e
                    }
                })
            }
        },
        localMethod(next, method) {
            return (...args) => {
                const start = Date.now()

                const result = next(...args)
                if (isPromise(result)) {
                    return result.then(data => {
                        const responseTime = Date.now() - start
                        console.log(traceWrapper(responseTime, method.name, 'method', args, start))
                        return data
                    })
                }
                const responseTime = Date.now() - start
                console.log(traceWrapper(responseTime, method.name, 'method', args, start))
                return result
            }
        },
    },
    logger
}